#include<Servo.h>
#include<LiquidCrystal_I2C.h>

// Erstelle Motorobjekte für die verbauten Servomotoren
Servo motorCircle;                  // Motor zum Drehen der beweglichen Drehscheibe
Servo motorMoveBall;                // Motor der die Kugel auf der Kupferbahn vom motorBallholder Servo zur Feder bewegt


// Pinverteilung
int PTPIN_LEFT = A1;                // Infrarottransistoren auf dem Spielbrett
int PTPIN_MIDDLE = A3;              // Infrarottransistoren auf dem Spielbrett
int PTPIN_RIGHT = A5;               // Infrarottransistoren auf dem Spielbrett
int PTPIN_START = A14;              // Infrarottransistor hinter den Flipperarmen
int PTPIN_COIN = A7;                // Infrarottransistor im Münzeinwurf
int LEDPIN_COIN = 8;                // Infrarotlampe im Münzeinwurf

int LED_YELLOW = 9;                 // Ansteuerung der gelben LEDs
int LED_COLOURFUL = 10;             // Ansteuerung der bunten LEDs
int SPEAKER_PIN = 11;               // Ansteuerung des Latsprechers

int ES_SWITCH_PIN = A9;             // Erschütterungssensor zum Tauschen der Flipperarme
int ES_CIRCLE_PIN = A11;            // Erschütterungssensor zum Drehen der beweglichen Drehscheibe
int ES_BLOCK_PIN = A13;             // Erschütterungssensor zum Blockieren der Flipperarme
int MOVEBALLPIN = 7;                // Ansteuerung des motorMoveBall Servos
int MOVECIRCLEPIN = 5;              // Ansteuerung des motorCircle Servos
int COMMUNICATIONPIN_SWITCH = 3;    // Kommunikation mit 2. Arduino (Nano) zum Auslösen des Vertauschens der Flipperarme
int COMMUNICATIONPIN_BLOCK = 2;     // Kommunikation mit 2. Arduino (Nano) zum Auslösen des Blockierens der Flipperarme


//Anlegen globaler Variablen
int positionMotorCircle = 0;        // speichert Drehwinkels der drehbaren Spielfläche
int score = 0;                      // speichert den aktuellen Punktestand
int highscore = 0;                  // speichert den Highscore spielübergreifend
int referenzValuePtMiddle;          // speichert Referenzwert für den mittleren Fototransistor
int referenzValuePtLeft;            // speichert Referenzwert für den mittleren Fototransistor
int referenzValuePtRight;           // speichert Referenzwert für den mittleren Fototransistor
int referenzValuePtStart;           // speichert Referenzwert für Fototransistor hinter den Flipperarmen

// Tonfrequenzen für die einzelnen Musiktöne
int g = 196;
int a = 220;
int h = 247;
int c1 = 262;
int d1 = 294;
int e1 = 330;
int f1 = 349;
int g1 = 392;
int a1 = 440;
int h1 = 494;
int e2 = 659;

//LCD Bildschirm deklariereen
LiquidCrystal_I2C lcd (0x27, 20, 4);

void setup() {

  //LCD Bildschirm konfigurieren
  lcd.begin();
  lcd.backlight();

  // Output und INPUT Zuweisung der einzelnen PINs
  pinMode(PTPIN_LEFT, INPUT);
  pinMode(PTPIN_MIDDLE, INPUT);
  pinMode(PTPIN_RIGHT, INPUT);
  pinMode(PTPIN_START, INPUT);
  pinMode(PTPIN_COIN, INPUT);
  pinMode(LEDPIN_COIN, OUTPUT);

  pinMode(LED_YELLOW, OUTPUT);
  pinMode(LED_COLOURFUL, OUTPUT);
  pinMode(SPEAKER_PIN, OUTPUT);

  pinMode(ES_SWITCH_PIN, INPUT);
  pinMode(ES_CIRCLE_PIN, INPUT);
  pinMode(ES_BLOCK_PIN, INPUT);
  pinMode(MOVEBALLPIN, OUTPUT);
  pinMode(MOVECIRCLEPIN, OUTPUT);
  pinMode(COMMUNICATIONPIN_SWITCH, OUTPUT);
  pinMode(COMMUNICATIONPIN_BLOCK, OUTPUT);

  // Motoren wird Pin zugewiesen
  motorMoveBall.attach(MOVEBALLPIN);
  motorCircle.attach(MOVECIRCLEPIN);

}

void loop() {

  // Start wird vorbereitet
  StartPosition();

  // Das Spiel wird gestartet
  lcd.setCursor(0, 0);                     // Dem LCD wird gezeigt, von wo aus er schreiben soll (erste Spalte, erste Zeile)
  lcd.print("Willkommen");
  lcd.setCursor(0, 1);                     // Dem LCD wird gezeigt, von wo aus er schreiben soll (erste Spalte, Zweite Zeile)
  delay(300);
  lcd.print("Bitte werfen Sie");
  lcd.setCursor(0, 2);                     // Dem LCD wird gezeigt, von wo aus er schreiben soll (erste Spalte, Zweite Zeile)
  delay(30);
  lcd.print("eine Muenze ein");
  
  /* Wenn die IR-Lampe im Münzeinwurf von einer Münze verdeckt wird, sinkt der gemessene Wert am dazugehörigen Phototransistor deutlich
     und das eigentliche Spiel beginnt.
     Solange wie jedoch keine Münze eingeworfen wird, passiert nichts und die while-Schleife wird immer wieder aufgerufen.
  */
  while (analogRead(PTPIN_COIN) > 1000) {
  }

  lcd.clear();                          // LCD wird geleert
  lcd.setCursor(0, 1);                  // Dem LCD wird gezeigt, von wo aus er schreiben soll (erste Spalte, Zweite Zeile)
  lcd.print("Das Spiel Beginnt.");
  delay(1000);
  lcd.clear();                          // LCD wird geleert

  // Das Spiel besteht aus 3 Druchläufen mit je einer Kugel. Jeder Durchgang wird durch einen Durchlauf der for-SChleife repräsentiert
  for (int i = 0; i < 3; i++) {
    // der Ball wird mit Hilfe der StarBall()-Fuktion ins Spielfeld gebracht
    StartBall();

    // manuelle Betätigung der Startvorrichtung nötig -> weshalb der Spieler aufgefordert wird, diese zu betätigen
    lcd.print("Bitte betaetigen Sie");
    lcd.setCursor(0, 2);
    delay(30);
    lcd.print("die Startvorrichtung"); // An dieser Stelle muss die Feder manuell gespannt und gelöst werden
    delay(1000);
    lcd.clear();
    lcd.print("Start");

    /* Solange bis der Ball hinter die Flipperarme fällt läuft der aktuelle Spieldurchgang (Durchlauf der for-Schleife)
       d.h. die Flipperarme können bewegt werden, Punkte können erzielt werden und die Sensoren ermitteln die Position der Kugel
       Falls jedoch eine Kugel vom Phototransistor hinter den Flipperarmen erfasst wird,
       wird sowohl die do-while SChleife, als auch der aktuelle Durchlauf in der for-schleife unterbrochen,
    */
    while (analogRead(PTPIN_START) >= referenzValuePtStart) {
      /* Einer der Spielfeldphototransistoren löst aus, wenn der Wert deutlich geringer als der Referenzwert ist.
         Es werden Punkte vergeben und eine Melodie abgespielt.
      */
      if (analogRead(PTPIN_RIGHT) < referenzValuePtRight) {
        AddPoints(20);                                      // Es werden Punkte vergeben und eine Melodie abgespielt.
      } else if (analogRead(PTPIN_MIDDLE) < referenzValuePtMiddle) {
        AddPoints(20);                                      // Es werden Punkte vergeben und eine Melodie abgespielt.
      } else if (analogRead(PTPIN_LEFT) < referenzValuePtLeft) {
        AddPoints(20);                                      // Es werden Punkte vergeben und eine Melodie abgespielt.

        //Wenn der Erschütterungssensor auslöst, sinkt der gemessene Widerstand.
      } else if (analogRead(ES_SWITCH_PIN) < 100) {
        AddPoints(30);                                    // Es werden Punkte vergeben und eine Melodie abgespielt.
        CommunicationWithNano(COMMUNICATIONPIN_SWITCH);  // Desweiteren wird ein Signal an den Nano gesendet, damit die Flipperarme vertauscht werden

        //Wenn der Erschütterungssensor auslöst, steigt die gemessene Spannung.
      } else if (analogRead(ES_BLOCK_PIN) < 100) {
        AddPoints(40);                                    // Es werden Punkte vergeben und eine Melodie abgespielt.
        CommunicationWithNano(COMMUNICATIONPIN_BLOCK);    // Desweiteren wird ein Signal an den Nano gesendet, damit die Flipperarme blockiert werden

        //Wenn der Erschütterungssensor auslöst, steigt die gemessene Spannung.
      } else if (analogRead(ES_CIRCLE_PIN) < 100) {
        AddPoints(10);                                    // Es werden Punkte vergeben und eine Melodie abgespielt.
        TurnPlayarea();                                   // Desweiteren dreht sich der drehbare Spielfeldteil.
      }
    }                                                      // Ende des aktuellen Durchlaufes, da eine Kugel hinter die Flipperarme gerollt ist.

    SoundtrackOfMario();                                  // Aufruf der Funktion zum Erzeugen von Licht-und Soundeffekten
    DisplayScore();                                       // Aufruf der Funktion zum Anzeigen des Punktestandes auf dem Display
    lcd.clear();                                          // LCD wird geleert
    delay(30);
  }                                                       // Alle 3 Durchläufe wurden gespielt. Das Spiel ist nun beendet.

  // Die erzielte Punktzahl wird angegeben
  lcd.clear();                                            // LCD wird geleert
  lcd.setCursor(5, 0);                                    // Dem LCD wird gezeigt, von wo aus er schreiben soll (sechste Spalte, erste Zeile)
  lcd.print("Game over!");
  delay(2000);
  lcd.setCursor(0, 1);                                    // Dem LCD wird gezeigt, von wo aus er schreiben soll (vierte Spalte, Zweite Zeile)
  lcd.print("Punkte:  " + (String) score);
  
  // kleine Besonderheit, falls der Highscore gebrochen wurde
  if (score > highscore) {
    SountrackOfTetris();                                 // Tetris Melodie wird gespielt falls der Highscore geknackt wurde
    lcd.setCursor(1, 0);                                 // Dem LCD wird gezeigt, von wo aus er schreiben soll (zweite Spalte, dritte Zeile)
    lcd.print("Neuer Highscore: ");
    delay(1000);
  }

  // Zurücksetzen des Spiels in den Ausgangszustand für neuen Spieldurchlauf
  StartPosition();
  delay(10000);
}


// Funktion setzt alle Elemente des Flippers in ihre Ausgangsposition, damit das Spiel (neu) gestartet werden kann
void StartPosition() {
  score = 0;
  motorCircle.write(positionMotorCircle);               // Die drehbare Spielfläche wird in  Ausgangsposition gebracht
  motorMoveBall.write(180);                             // Die Vorrichtung zum Bewegen der Kugel zur Startvorrichtung wird in Ausgangsposition gebracht
  digitalWrite(LEDPIN_COIN, HIGH);                      // Die LED im Münzeinwurf wird angeschaltet (Ausgangsposition der Lichtschranke)
  digitalWrite(COMMUNICATIONPIN_SWITCH, LOW);           // Der Pin zur Kommunikation mit dem Arduino zum Tauschen der Flipperarme wird ausgeschaltet (Ausgangsposition)
  digitalWrite(COMMUNICATIONPIN_BLOCK, LOW);            // Der Pin zur Kommunikation mit dem Arduino zum Blockieren der Flipperarme wird ausgeschaltet (Ausgangsposition)
  referenzValuePtLeft = analogRead(PTPIN_LEFT) - 9;
  referenzValuePtMiddle = analogRead(PTPIN_MIDDLE) - 80;
  referenzValuePtRight = analogRead(PTPIN_RIGHT) - 100;
  referenzValuePtStart = analogRead(PTPIN_START) - 80;
  lcd.clear();
}

// Funktion zum Einleiten des Startes: Kugel wird mit Hilfe des moveBall Motors zur Startvorrichtung transportiert
void StartBall() {                                      // Kugel wird bis zur Startvorrichtung geschoben
  delay(1500);
  for (int i = 180; i > 40; i--) {                      // zurück in Ausgangsposition
    motorMoveBall.write(i);
    delay(5);
  }
  delay(1000);
  motorMoveBall.write(180);
}

// Funktion zum Addieren von erziehlten Punkten zum aktuellen Punktestand
void AddPoints(int points) {
  score += points;                                     // Punkte werden dem Punktestand hinzugefügt
  SoundtrackOfMarioCoin();                             // spielt den Coin Sound von Super Mario mit entsprechenden Lichteffekten
  DisplayScore();                                      // Funktion zum Ausgeben der Punktzahl wird aufgerufen
}

// Funktion zum Darstellen der Punktzahl auf dem Display
void DisplayScore() {
  lcd.clear();                                         // LCD wird geleert
  lcd.setCursor(7, 1);                                 // Dem LCD wird gezeigt, von wo aus er schreiben soll (erste Spalte, Zweite Zeile)
  lcd.print(score);                                    // Audgabe der Punktzahl
}

/* Falls einer der Erschütterungssensoren des Arduino Mega ausgelöst wird, wird an dem entsprechenden Pins des Arduino Megas eine Spannung erzeugt,
   welche von Arduino Nano registriert wird, sodass dieser den dazugehörigen Spezialeffekt einleiten kann
*/
  void CommunicationWithNano(int pin) {
  digitalWrite(pin, HIGH);
  delay(30);
  digitalWrite(pin, LOW);
  }

// Funktion zum Drehen der drehbaren Spielfeldscheibe
void TurnPlayarea() {
  // falls der Drehwinkel 0° beträgt, wird die Drehscheibe auf 120° gedreht
  if ( positionMotorCircle == 0) {
    motorCircle.write(120);

    // falls der Drehwinkel bereits  120° beträgt, wird die Drehscheibe zurück  auf 0°gedreht
  } else {
    motorCircle.write(0);
  }
}

/* Ein Ton wird mit Hilfe der übergebenen Werte erzeugt (Tonhöhe/ Frequenz, Tonlänge)
   passend zu dem Ton leuchten die übergebenen LEDs auf und der aktuelle Punktestand erscheint auf dem Display
*/
void playMusic(int ledPin, int musicalNote, int toneLength) {
  lcd.clear();
  lcd.setCursor(7, 1);
  lcd.print(score);
  tone(SPEAKER_PIN, musicalNote, toneLength);     // Ton wird erzeugt
  digitalWrite(ledPin, HIGH);                     // LED leuchtet auf
  delay(toneLength - 5);
  digitalWrite(ledPin, LOW);                      // LED geht aus
}


/* Funktion spielt den Coin Sound von Super Mario, indem sie die playMusic()-Funktion immerwieder aufruft
    und ihr die entsprechenden Tonfrequenzen, Tonlengen und abwecheslnd eine der LED-Pins übergibt
*/
void SoundtrackOfMarioCoin() {
  playMusic(LED_COLOURFUL, h1, 125);
  playMusic(LED_YELLOW, e2, 375);
  delay(500);
}


/* Funktion spielt die Melodie von Super Mario, indem sie die playMusic()-Funktion immerwieder aufruft
    und ihr die entsprechenden Tonfrequenzen, Tonlengen und abwecheslnd eine der LED-Pins übergibt
*/
void SoundtrackOfMario() {
  playMusic(LED_COLOURFUL, e1, 125);
  playMusic(LED_YELLOW, e1, 125);
  delay(125);
  playMusic(LED_COLOURFUL, e1, 125);
  delay(125);
  playMusic(LED_YELLOW, c1, 125);
  playMusic(LED_COLOURFUL, e1, 250);
  playMusic(LED_YELLOW, g1, 250);
  delay(250);
  playMusic(LED_COLOURFUL, g, 500);
  delay(250);
}


/* Funktion spielt die Melodie von Tetris, indem sie die playMusic()-Funktion immer wieder aufruft
    und ihr die entsprechenden Tonfrequenzen, Tonlängen und abwecheslnd eine der LED Pins übergibt
*/
void SountrackOfTetris() {
  playMusic(LED_YELLOW, e1, 300);
  playMusic(LED_COLOURFUL, h, 150);
  playMusic(LED_YELLOW, c1, 150);
  playMusic(LED_COLOURFUL, d1, 300);
  playMusic(LED_YELLOW, c1, 150);
  playMusic(LED_COLOURFUL, h, 150);

  playMusic(LED_YELLOW, a, 300);
  playMusic(LED_COLOURFUL, a, 150);
  playMusic(LED_YELLOW, c1, 150);
  playMusic(LED_COLOURFUL, e1, 300);
  playMusic(LED_YELLOW, d1, 150);
  playMusic(LED_COLOURFUL, c1, 150);

  playMusic(LED_YELLOW, h, 300);
  playMusic(LED_COLOURFUL, h, 150);
  playMusic(LED_YELLOW, c1, 150);
  playMusic(LED_COLOURFUL, d1, 300);
  playMusic(LED_YELLOW, e1, 300);

  playMusic(LED_COLOURFUL, c1, 300);
  playMusic(LED_YELLOW, a, 300);
  playMusic(LED_COLOURFUL, a, 300);

  delay(300);
  playMusic(LED_YELLOW, d1, 300);
  playMusic(LED_COLOURFUL, d1, 150);
  playMusic(LED_YELLOW, f1, 150);
  playMusic(LED_COLOURFUL, a1, 300);
  playMusic(LED_YELLOW, g1, 150);
  playMusic(LED_COLOURFUL, f1, 150);

  playMusic(LED_YELLOW, e1, 300);
  playMusic(LED_COLOURFUL, e1, 150);
  playMusic(LED_YELLOW, c1, 150);
  playMusic(LED_COLOURFUL, e1, 300);
  playMusic(LED_YELLOW, d1, 150);
  playMusic(LED_COLOURFUL, c1, 150);

  playMusic(LED_YELLOW, h, 300);
  playMusic(LED_COLOURFUL, h, 150);
  playMusic(LED_YELLOW, c1, 150);
  playMusic(LED_COLOURFUL, d1, 300);
  playMusic(LED_YELLOW, e1, 300);

  playMusic(LED_COLOURFUL, c1, 300);
  playMusic(LED_YELLOW, a, 300);
  playMusic(LED_COLOURFUL, a, 300);
  delay(300);
}
